///////////////////////////////////////////////////////////////////////////////
///
/// Written 2012, Oliver Schneider (assarbad.net) - PUBLIC DOMAIN/CC0
///
/// Purpose          : Class to get the Windows(tm) product key from the
///                    registry.
///
///////////////////////////////////////////////////////////////////////////////
#ifndef __WINPRODUCTKEY_H_VER__
#define __WINPRODUCTKEY_H_VER__ 2012120401
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif // Check for "#pragma once" support

#include <cassert>
#include <tchar.h>
#include <Windows.h>

// DDK build fails otherwise on older 2003 Server DDK
#ifndef KEY_WOW64_32KEY
#define KEY_WOW64_32KEY         (0x0200)
#endif // KEY_WOW64_32KEY
#ifndef KEY_WOW64_64KEY
#define KEY_WOW64_64KEY         (0x0100)
#endif // KEY_WOW64_64KEY


namespace
{
    LPCTSTR DefaultKeyName = _T("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion");
    LPCTSTR DefaultMSDPIDName = _T("DigitalProductId");
}

#ifdef _DEBUG
#   define DBGPRINT 0
#   define dbgprint(fmt, ...)   if(DBGPRINT) do { \
                                    _ftprintf(stderr, _T("DBG: ")); \
                                    _ftprintf(stderr, fmt, __VA_ARGS__); \
                                    _ftprintf(stderr, _T("\n")); \
                                    fflush(stderr); \
                                } while(0)
#else
#   define dbgprint /##/
#endif // _DEBUG

class CWindowsProductKey
{
public:
    // Allow the header to be included again outside of this file
    // Still this header needs to be included before anyone else who requires VerySimpleBuf.h
#   define __VERYSIMPLEBUF_MULTI_INC__
#   include "VerySimpleBuf.h"
#   undef __VERYSIMPLEBUF_H_VER__
    // Just make it more readable
    typedef CVerySimpleBuf<WCHAR> CWideString, *PWideString;
    typedef CVerySimpleBuf<CHAR>  CAnsiString, *PAnsiString;
    typedef CVerySimpleBuf<TCHAR> CTString, *PTString;
    typedef CVerySimpleBuf<BYTE>  ByteBuf;

private:
    mutable LONG m_lastError;
    DWORD m_fakeBuildNo;
    LPCTSTR m_keyName;
    LPCTSTR m_MSDPIDName;
    DWORD m_MSDPIDLen;
    ByteBuf m_MSDPIDValue;
    bool m_readRegOK;
    CTString m_decodedKey;
public:
    // Default ctor
    CWindowsProductKey(
            HKEY rootKey = HKEY_LOCAL_MACHINE
            , LPCTSTR szKeyName = DefaultKeyName
            , LPCTSTR szMSDPIDName = DefaultMSDPIDName
            , DWORD winFakeBuildNo = -1
        )
        : m_lastError(ERROR_SUCCESS)
        , m_fakeBuildNo(winFakeBuildNo)
        , m_keyName(szKeyName)
        , m_MSDPIDName(szMSDPIDName)
        , m_MSDPIDLen(0)
        , m_readRegOK(ctorReadRegValues_(rootKey))
        , m_decodedKey(decodeMSDPID_<TCHAR>())
    {
    }

    CWindowsProductKey(LPCTSTR szMSDPIDName, DWORD winFakeBuildNo = -1)
        : m_lastError(ERROR_SUCCESS)
        , m_fakeBuildNo(winFakeBuildNo)
        , m_keyName(DefaultKeyName)
        , m_MSDPIDName(szMSDPIDName)
        , m_MSDPIDLen(0)
        , m_readRegOK(ctorReadRegValues_(HKEY_LOCAL_MACHINE))
        , m_decodedKey(decodeMSDPID_<TCHAR>())
    {
    }

    CWindowsProductKey(const BYTE* dpidBuf, DWORD dpidBufLen, DWORD winFakeBuildNo = -1)
        : m_lastError(ERROR_SUCCESS)
        , m_fakeBuildNo(winFakeBuildNo)
        , m_keyName(DefaultKeyName)
        , m_MSDPIDName(NULL)
        , m_MSDPIDLen(0)
        , m_readRegOK(ctorCopyBuffer_(dpidBuf, dpidBufLen))
        , m_decodedKey(decodeMSDPID_<TCHAR>())
    {
    }

    LPCTSTR decodedKey() const
    {
        return m_decodedKey.getBuf();
    }

    LONG lastError() const
    {
        return m_lastError;
    }

private:
    static const size_t dpidOffs3 = 52;
    static const size_t dpidOffsX = 808;
    static const size_t dpidLenB  = 14;

    template<typename T> bool readRegValue_(HKEY hKey, LPCTSTR szValueName, CVerySimpleBuf<T>& buf, DWORD& cbBufLen, DWORD expectType = REG_NONE)
    {
        DWORD dwRegType = REG_NONE;
        if(ERROR_SUCCESS == (m_lastError = ::RegQueryValueEx(hKey, szValueName, 0, &dwRegType, NULL, &cbBufLen)))
        {
            if(expectType != REG_NONE && dwRegType != expectType)
            {
                // Enable REG_SZ to REG_EXPAND_SZ compatibility
                if((REG_EXPAND_SZ == expectType) || (REG_SZ == expectType))
                {
                    if((REG_EXPAND_SZ == expectType) && (REG_SZ != dwRegType))
                    {
                        return false;
                    }
                    if((REG_SZ == expectType) && (REG_EXPAND_SZ != dwRegType))
                    {
                        return false;
                    }
                }
                else
                    return false;
            }
            DWORD toAlloc = cbBufLen;
            if((REG_EXPAND_SZ == dwRegType)  && (REG_SZ == dwRegType))
            {
                toAlloc = cbBufLen / sizeof(TCHAR) + 1; // accommodate zero-terminator
            }
            if(!buf.reAlloc(toAlloc))
            {
                m_lastError = ERROR_NOT_ENOUGH_MEMORY;
                return false;
            }
            if(ERROR_SUCCESS == (m_lastError = ::RegQueryValueEx(hKey, szValueName, 0, &dwRegType, (LPBYTE)buf.getBuf(), &cbBufLen)))
            {
                return true;
            }
        }
        return false;
    }

    template<typename T> CVerySimpleBuf<T> decodeMSDPID_()
    {
        // Check the environment
        if(!m_readRegOK || !m_MSDPIDValue)
        {
            if(!m_readRegOK)
                dbgprint(_T("Registry values not read"));
            if(!m_MSDPIDValue)
                dbgprint(_T("Binary value of MSDPID is empty"));
            return size_t(0);
        }

        const BYTE* buf = m_MSDPIDValue.getBuf();
        const WORD wPIDVer = (buf[5] << 8) | buf[4];
        const size_t dpidOffs = (wPIDVer == 3) ? dpidOffs3 : dpidOffsX;
        // Another sanity check
        if(m_MSDPIDLen <= (dpidOffs + dpidLenB))
        {
            dbgprint(_T("Binary value too short for meaningful MSDPID (%u bytes, %u expected)"), m_MSDPIDLen, dpidOffs + dpidLenB);
            return size_t(0);
        }

        CHAR pkStr[30] = {0};

        // Decode, callee expects valid arguments
        if(!decodeRawMSDPID_(buf, m_MSDPIDLen, pkStr, sizeof(pkStr), dpidOffs, m_fakeBuildNo))
        {
            dbgprint(_T("Failure to decode MSDPID. Out of memory?"));
            return size_t(0);
        }

        dbgprint(_T("Decoded key: %hs"), pkStr);
        CVerySimpleBuf<T> retstr(sizeof(pkStr));
        // Copy over (looks complicated, but exists so the template works for TCHAR)
        if(T* strbuf = retstr.getBuf())
            for(size_t i = 0; i < sizeof(pkStr); i++)
                strbuf[i] = (T)pkStr[i];
        return retstr;
    }

    static void transposeKey_(CHAR* buf, size_t len, bool bWin8, DWORD winBuildNum)
    {
        WORD wBuildNum = (winBuildNum < 0x10000) ? (WORD)winBuildNum : (WORD)(::GetVersion() >> 16);
        // Special treatment for Windows 8 and higher
        if(bWin8 || wBuildNum > 7601)
        {
            memmove(buf, buf+1, 2);
            buf[2] = 'N';
        }
    }

    static bool decodeRawMSDPID_(const BYTE* buf, size_t bufLen, CHAR* pkStr, size_t pkStrLen, const size_t dpidOffs, DWORD winBuildNum)
    {
        assert(pkStrLen == 30);
        const CHAR pkCS[] = "BCDFGHJKMPQRTVWXY2346789";
        const size_t pkCSLen = sizeof(pkCS)-1; // = 24

        ByteBuf dpidBuf(bufLen); // take copy, we'll modify it ...
        if(!dpidBuf)
            return false;
        BYTE* dpid = dpidBuf.getBuf(); // alias

        memset(pkStr, 0, pkStrLen);
        memcpy(dpid, buf, bufLen);
        dpid += dpidOffs; // skip forward to the relevant part of the buffer

        const DWORD* pdwDpidBuf = (PDWORD)buf;
        if(!pdwDpidBuf[2]) // sanity check
            return false;

        const BYTE MaskedEndByte = (dpid[dpidLenB] & 0xF7);
        const bool bWin8 = (dpid[dpidLenB] != MaskedEndByte);
        dpid[dpidLenB] = MaskedEndByte;

        for(int i = dpidLenB*2; i >= 0; i--)
        {
            DWORD dwAcc = 0;
            for(int j = dpidLenB; j >= 0; j--)
            {
                dwAcc *= 0x100;
                dwAcc += dpid[j];
                dpid[j] = (BYTE)(((dwAcc / pkCSLen) < 0x100) ? (dwAcc / pkCSLen) : 0xFF);
                dwAcc %= pkCSLen;
            }
            pkStr[i] = pkCS[dwAcc];
            if(i && 0 == (pkStrLen-i) % 6)
            {
                pkStr[--i] = '-';
            }
        }
        transposeKey_(pkStr, pkStrLen, bWin8, winBuildNum);
        return true;
    }

    bool ctorReadRegValues_(HKEY rootKey)
    {
        bool retval = false;
        HKEY hKey = NULL;
        m_lastError = ::RegOpenKeyEx(rootKey, m_keyName, 0, KEY_QUERY_VALUE | KEY_WOW64_64KEY, &hKey);
        if(ERROR_SUCCESS == m_lastError)
        {
            dbgprint(_T("[0x%08X]\\%s opened as %p"), rootKey, m_keyName, hKey);
            DWORD dwRegType = REG_NONE;

            if(readRegValue_(hKey, m_MSDPIDName, m_MSDPIDValue, m_MSDPIDLen, REG_BINARY))
            {
                dbgprint(_T("%s of length %u read"), m_MSDPIDName, m_MSDPIDLen);
                retval = true;
                dbgprint(_T("Returning success"));
            }
            ::RegCloseKey(hKey);
        }
        return retval;
    }

    bool ctorCopyBuffer_(const BYTE* buf, DWORD len)
    {
        if(m_MSDPIDValue.reAlloc(len))
        {
            memcpy(m_MSDPIDValue.getBuf(), buf, len);
            m_MSDPIDLen = len;
            dbgprint(_T("Copied buffer of length %u"), m_MSDPIDLen);
            return true;
        }
        dbgprint(_T("Failed to allocate buffer for MSDPID"));
        return false;
    }

    CWindowsProductKey(const CWindowsProductKey&);
    CWindowsProductKey& operator=(const CWindowsProductKey&);
};

#endif // __WINPRODUCTKEY_H_VER__
